@@ -4,6 +4,8 @@ import (
 	"bufio"
 	"io/ioutil"
 	"os"
+	"path/filepath"
+	"reflect"
 	"regexp"
 	"strconv"
 	"strings"
@@ -12,6 +14,10 @@ import (
 	"gopkg.in/yaml.v2"
 )
 
+var (
+	reModuledir = regexp.MustCompile(`^\s*(?:moduledir)\s+['\"]?([^'\"]+)['\"]?`)
+)
+
 // readConfigfile creates the ConfigSettings struct from the g10k config file
 func readConfigfile(configFile string) ConfigSettings {
 	Debugf("Trying to read g10k config file: " + configFile)
@@ -22,7 +28,7 @@ func readConfigfile(configFile string) ConfigSettings {
 
 	rubySymbolsRemoved := ""
 	for _, line := range strings.Split(string(data), "\n") {
-		reWhitespaceColon := regexp.MustCompile("^(\\s*):")
+		reWhitespaceColon := regexp.MustCompile(`^(\s*):`)
 		m := reWhitespaceColon.FindStringSubmatch(line)
 		if len(m) > 0 {
 			rubySymbolsRemoved += reWhitespaceColon.ReplaceAllString(line, m[1]) + "\n"
@@ -36,9 +42,6 @@ func readConfigfile(configFile string) ConfigSettings {
 		Fatalf("YAML unmarshal error: " + err.Error())
 	}
 
-	//fmt.Print("config: ")
-	//fmt.Printf("%+v\n", config)
-
 	if len(os.Getenv("g10k_cachedir")) > 0 {
 		cachedir := os.Getenv("g10k_cachedir")
 		Debugf("Found environment variable g10k_cachedir set to: " + cachedir)
@@ -48,15 +51,15 @@ func readConfigfile(configFile string) ConfigSettings {
 	}
 
 	config.CacheDir = checkDirAndCreate(config.CacheDir, "cachedir")
-	config.ForgeCacheDir = checkDirAndCreate(config.CacheDir+"forge/", "cachedir/forge")
-	config.ModulesCacheDir = checkDirAndCreate(config.CacheDir+"modules/", "cachedir/modules")
-	config.EnvCacheDir = checkDirAndCreate(config.CacheDir+"environments/", "cachedir/environments")
+	config.ForgeCacheDir = checkDirAndCreate(filepath.Join(config.CacheDir, "forge"), "cachedir/forge")
+	config.ModulesCacheDir = checkDirAndCreate(filepath.Join(config.CacheDir, "modules"), "cachedir/modules")
+	config.EnvCacheDir = checkDirAndCreate(filepath.Join(config.CacheDir, "environments"), "cachedir/environments")
 
-	if len(config.Forge.Baseurl) == 0 {
-		config.Forge.Baseurl = "https://forgeapi.puppetlabs.com"
+	if len(config.ForgeBaseURL) == 0 {
+		config.ForgeBaseURL = "https://forgeapi.puppet.com"
 	}
 
-	//fmt.Println("Forge Baseurl: ", config.Forge.Baseurl)
+	// fmt.Println("Forge Baseurl: ", config.ForgeBaseURL)
 
 	// set default timeout to 5 seconds if no timeout setting found
 	if config.Timeout == 0 {
@@ -99,10 +102,49 @@ func readConfigfile(configFile string) ConfigSettings {
 		config.MaxExtractworker = 20
 	}
 
+	if len(config.ForgeCacheTTLString) != 0 {
+		ttl, err := time.ParseDuration(config.ForgeCacheTTLString)
+		if err != nil {
+			Fatalf("Error: Can not convert value " + config.ForgeCacheTTLString + " of config setting forge_cache_ttl to a golang Duration. Valid time units are 300ms, 1.5h or 2h45m. In " + configFile)
+		}
+		config.ForgeCacheTTL = ttl
+	}
+
+	// check for non-empty config.Deploy which takes precedence over the non-deploy scoped settings
+	// See https://github.com/puppetlabs/r10k/blob/master/doc/dynamic-environments/configuration.mkd#deploy
+	emptyDeploy := DeploySettings{}
+	if !reflect.DeepEqual(config.Deploy, emptyDeploy) {
+		Debugf("detected deploy configuration hash, which takes precedence over the non-deploy scoped settings")
+		config.PurgeLevels = config.Deploy.PurgeLevels
+		config.PurgeAllowList = config.Deploy.PurgeAllowList
+		config.DeploymentPurgeAllowList = config.Deploy.DeploymentPurgeAllowList
+		config.WriteLock = config.Deploy.WriteLock
+		config.GenerateTypes = config.Deploy.GenerateTypes
+		config.PuppetPath = config.Deploy.PuppetPath
+		config.PurgeSkiplist = config.Deploy.PurgeSkiplist
+		config.Deploy = emptyDeploy
+	}
+
+	if len(config.PurgeLevels) == 0 {
+		config.PurgeLevels = []string{"deployment", "puppetfile"}
+	}
+
+	for source, sa := range config.Sources {
+		sa.Basedir = normalizeDir(sa.Basedir)
+
+		// set default to "correct_and_warn" like r10k
+		// https://github.com/puppetlabs/r10k/blob/master/doc/dynamic-environments/git-environments.mkd#invalid_branches
+		if len(sa.AutoCorrectEnvironmentNames) == 0 {
+			sa.AutoCorrectEnvironmentNames = "correct_and_warn"
+		}
+		config.Sources[source] = sa
+	}
+
 	if validate {
 		Validatef()
 	}
 
+	// fmt.Printf("%+v\n", config)
 	return config
 }
 
@@ -114,8 +156,8 @@ func preparePuppetfile(pf string) string {
 	}
 	defer file.Close()
 
-	reComma := regexp.MustCompile(",\\s*$")
-	reComment := regexp.MustCompile("^\\s*#")
+	reComma := regexp.MustCompile(`,\s*$`)
+	reComment := regexp.MustCompile(`^\s*#`)
 	reEmpty := regexp.MustCompile("^$")
 
 	pfString := ""
@@ -144,7 +186,7 @@ func preparePuppetfile(pf string) string {
 }
 
 // readPuppetfile creates the ConfigSettings struct from the Puppetfile
-func readPuppetfile(pf string, sshKey string, source string, forceForgeVersions bool, replacedPuppetfileContent bool) Puppetfile {
+func readPuppetfile(pf string, sshKey string, source string, branch string, forceForgeVersions bool, replacedPuppetfileContent bool) Puppetfile {
 	var puppetFile Puppetfile
 	var n string
 	puppetFile.privateKey = sshKey
@@ -159,17 +201,16 @@ func readPuppetfile(pf string, sshKey string, source string, forceForgeVersions
 		n = preparePuppetfile(pf)
 	}
 
-	reEmptyLine := regexp.MustCompile("^\\s*$")
-	reModuledir := regexp.MustCompile("^\\s*(?:moduledir)\\s+['\"]?([^'\"]+)['\"]?")
-	reForgeCacheTTL := regexp.MustCompile("^\\s*(?:forge.cache(?:TTL|Ttl))\\s+['\"]?([^'\"]+)['\"]?")
-	reForgeBaseURL := regexp.MustCompile("^\\s*(?:forge.base(?:URL|Url))\\s+['\"]?([^'\"]+)['\"]?")
-	reForgeModule := regexp.MustCompile("^\\s*(?:mod)\\s+['\"]?([^'\"]+[-/][^'\"]+)['\"](?:\\s*)[,]?(.*)")
-	reForgeAttribute := regexp.MustCompile("\\s*['\"]?([^\\s'\"]+)\\s*['\"]?(?:=>)?\\s*['\"]?([^'\"]+)?")
-	reGitModule := regexp.MustCompile("^\\s*(?:mod)\\s+['\"]?([^'\"/]+)['\"]\\s*,(.*)")
-	reGitAttribute := regexp.MustCompile("\\s*:(git|commit|tag|branch|ref|link|ignore[-_]unreachable|fallback|install_path|default_branch|local)\\s*=>\\s*['\"]?([^'\"]+)['\"]?")
-	reUniqueGitAttribute := regexp.MustCompile("\\s*:(?:commit|tag|branch|ref|link)\\s*=>")
-	reDanglingAttribute := regexp.MustCompile("^\\s*:[^ ]+\\s*=>")
-	moduleDir := "modules/"
+	reEmptyLine := regexp.MustCompile(`^\s*$`)
+	reForgeCacheTTL := regexp.MustCompile(`^\s*(?:forge.cache(?:TTL|Ttl))\s+['\"]?([^'\"]+)['\"]?`)
+	reForgeBaseURL := regexp.MustCompile(`^\s*(?:forge.base(?:URL|Url))\s+['\"]?([^'\"]+)['\"]?`)
+	reForgeModule := regexp.MustCompile(`^\s*(?:mod)\s+['\"]?([^'\"]+[-/][^'\"]+)['\"](?:\s*)[,]?(.*)`)
+	reForgeAttribute := regexp.MustCompile(`\s*['\"]?([^\s'\"]+)\s*['\"]?(?:=>)?\s*['\"]?([^'\"]+)?`)
+	reGitModule := regexp.MustCompile(`^\s*(?:mod)\s+['\"]?([^'\"/]+)['\"]\s*,(.*)`)
+	reGitAttribute := regexp.MustCompile(`\s*:(git|commit|tag|branch|ref|link|ignore[-_]unreachable|fallback|install_path|default_branch|local|use_ssh_agent)\s*=>\s*['\"]?([^'\"]+)['\"]?`)
+	reUniqueGitAttribute := regexp.MustCompile(`\s*:(?:commit|tag|branch|ref|link)\s*=>`)
+	reDanglingAttribute := regexp.MustCompile(`^\s*:[^ ]+\s*=>`)
+	moduleDir := "modules"
 	var moduleDirs []string
 	//nextLineAttr := false
 
@@ -210,8 +251,10 @@ func readPuppetfile(pf string, sshKey string, source string, forceForgeVersions
 			forgeModuleName := strings.TrimSpace(m[1])
 			//fmt.Println("found forge mod name ------------------------------> ", forgeModuleName)
 			comp := strings.Split(forgeModuleName, "/")
+			forgeModuleNameSeparator := "/"
 			if len(comp) != 2 {
 				comp = strings.Split(forgeModuleName, "-")
+				forgeModuleNameSeparator = "-"
 				if len(comp) != 2 {
 					Fatalf("Error: Forge module name is invalid! Should be like puppetlabs/apt or puppetlabs-apt, but is: " + m[2] + " in " + pf + " line: " + line)
 				}
@@ -220,6 +263,7 @@ func readPuppetfile(pf string, sshKey string, source string, forceForgeVersions
 			if _, ok := puppetFile.forgeModules[comp[1]]; ok {
 				Fatalf("Error: Duplicate forge module found in " + pf + " for module " + forgeModuleName + " line: " + line)
 			}
+			//Debugf("Found Forge module name " + forgeModuleName + " with " + forgeModuleNameSeparator + " as a separator")
 			forgeModuleVersion := "present"
 			forgeChecksum := ""
 			// try to find a forge module attribute
@@ -247,17 +291,17 @@ func readPuppetfile(pf string, sshKey string, source string, forceForgeVersions
 							// try to detect Git modules in Forge <AUTHOR>/<MODULENAME> notation, fixes #104
 							Debugf("Found git module in Forge notation: " + forgeModuleName + " with git url: " + forgeAttributeValue)
 							//fmt.Println("line:", line)
-							removeForgeNotationAuthor := strings.Split(line, "/")
+							removeForgeNotationAuthor := strings.Split(line, forgeModuleNameSeparator)
 							if len(removeForgeNotationAuthor) < 2 {
 								Fatalf("Error: Found git module in Forge notation: " + forgeModuleName + " with git url: " + forgeAttributeValue + ", but something went wrong while trying to remove the author part to make g10k detect it as an Git module module:" + comp[1] + " line: " + line)
 							} else {
 								//fmt.Println("removeForgeNotationAuthor:", removeForgeNotationAuthor[0])
-								replacedLine := strings.Replace(line, removeForgeNotationAuthor[0]+"/", "mod '", 1)
+								replacedLine := strings.Replace(line, removeForgeNotationAuthor[0]+forgeModuleNameSeparator, "mod '", 1)
 								//fmt.Println("replacedLine:", replacedLine)
 								//fmt.Print("n:", n)
 								newN := strings.Replace(n, line, replacedLine, 1)
 								//fmt.Print("newN:", newN)
-								return readPuppetfile(newN, sshKey, source, forceForgeVersions, true)
+								return readPuppetfile(newN, sshKey, source, branch, forceForgeVersions, true)
 							}
 						}
 					}
@@ -269,7 +313,11 @@ func readPuppetfile(pf string, sshKey string, source string, forceForgeVersions
 			if _, ok := puppetFile.gitModules[comp[1]]; ok {
 				Fatalf("Error: Forge Puppet module with same name found in " + pf + " for module " + comp[1] + " line: " + line)
 			}
-			puppetFile.forgeModules[comp[1]] = ForgeModule{version: forgeModuleVersion, name: comp[1], author: comp[0], sha256sum: forgeChecksum, moduleDir: moduleDir}
+			// the base url in the Puppetfile takes precedence over an base url specified in the g10k config yaml
+			if len(puppetFile.forgeBaseURL) == 0 {
+				puppetFile.forgeBaseURL = config.ForgeBaseURL
+			}
+			puppetFile.forgeModules[comp[1]] = ForgeModule{version: forgeModuleVersion, name: comp[1], author: comp[0], sha256sum: forgeChecksum, moduleDir: moduleDir, sourceBranch: source + "_" + branch}
 		} else if m := reGitModule.FindStringSubmatch(line); len(m) > 1 {
 			gitModuleName := m[1]
 			//fmt.Println("found git mod name ---> ", gitModuleName)
@@ -320,12 +368,10 @@ func readPuppetfile(pf string, sshKey string, source string, forceForgeVersions
 						gm.git = a[2]
 					} else if gitModuleAttribute == "branch" {
 						if a[2] == ":control_branch" || a[2] == "control_branch" {
-							if strings.HasPrefix(a[2], ":") {
-								a[2] = strings.TrimLeft(a[2], ":")
-							}
 							gm.link = true
+						} else {
+							gm.branch = a[2]
 						}
-						gm.branch = a[2]
 					} else if gitModuleAttribute == "tag" {
 						gm.tag = a[2]
 					} else if gitModuleAttribute == "commit" {
@@ -361,6 +407,12 @@ func readPuppetfile(pf string, sshKey string, source string, forceForgeVersions
 						if local {
 							gm.local = true
 						}
+					} else if gitModuleAttribute == "use_ssh_agent" {
+						useSSHAgent, err := strconv.ParseBool(a[2])
+						if err != nil {
+							Fatalf("Error: Can not convert value " + a[2] + " of parameter " + gitModuleAttribute + " to boolean. In " + pf + " for module " + gitModuleName + " line: " + line)
+						}
+						gm.useSSHAgent = useSSHAgent
 					}
 
 				}
@@ -393,6 +445,7 @@ func readPuppetfile(pf string, sshKey string, source string, forceForgeVersions
 	}
 
 	puppetFile.moduleDirs = moduleDirs
-	//fmt.Println(puppetFile)
+	puppetFile.sourceBranch = branch
+	// fmt.Printf("%+v\n", puppetFile)
 	return puppetFile
 }
