@@ -1,7 +1,6 @@
 package main
 
 import (
-	"archive/tar"
 	"crypto/md5"
 	"crypto/sha256"
 	"encoding/hex"
@@ -23,10 +22,32 @@ import (
 	"github.com/xorpaul/uiprogress"
 )
 
+func checkDeprecation(fm ForgeModule, lastCheckedFile string) bool {
+	// check content of lastCheckedFile (which should be the Forge API response body) if the module is deprecated
+	// return false if the api needs to be queried again
+	if fileInfo, err := os.Stat(lastCheckedFile); err == nil {
+		if fileInfo.Size() < 1 {
+			Debugf("found empty file " + lastCheckedFile)
+			return false
+		} else if fm.cacheTTL > 0 && fileInfo.ModTime().Add(fm.cacheTTL).Before(time.Now()) {
+			return false
+		} else {
+			json, err := ioutil.ReadFile(lastCheckedFile)
+			if err != nil {
+				Fatalf("doModuleInstallOrNothing(): Error while reading Forge API result from file " + lastCheckedFile + err.Error())
+			}
+			_ = parseForgeAPIResult(string(json), fm)
+			return true
+		}
+	}
+	return false
+}
+
 func doModuleInstallOrNothing(fm ForgeModule) {
 	moduleName := fm.author + "-" + fm.name
 	moduleVersion := fm.version
-	workDir := config.ForgeCacheDir + moduleName + "-" + fm.version
+	workDir := filepath.Join(config.ForgeCacheDir, moduleName+"-"+fm.version)
+	lastCheckedFile := filepath.Join(config.ForgeCacheDir, moduleName+"-latest-last-checked")
 	fr := ForgeResult{false, fm.version, "", 0}
 	if check4update {
 		moduleVersion = "latest"
@@ -40,14 +61,14 @@ func doModuleInstallOrNothing(fm ForgeModule) {
 				if _, ok := uniqueForgeModules[moduleName+"-"+fr.versionNumber]; ok {
 					Debugf("no need to fetch Forge module " + moduleName + " in latest, because latest is " + fr.versionNumber + " and that will already be fetched")
 					fr.needToGet = false
-					versionDir := config.ForgeCacheDir + moduleName + "-" + fr.versionNumber
+					versionDir := filepath.Join(config.ForgeCacheDir, moduleName+"-"+fr.versionNumber)
 					absolutePath, err := filepath.Abs(versionDir)
 					Debugf("trying to create symlink " + workDir + " pointing to " + absolutePath)
 					if err != nil {
 						Fatalf("doModuleInstallOrNothing(): Error while resolving absolute file path for " + versionDir + " Error: " + err.Error())
 					}
 					if err := os.Symlink(absolutePath, workDir); err != nil {
-						Fatalf("doModuleInstallOrNothing(): 1 Error while creating symlink " + workDir + " pointing to " + absolutePath + err.Error())
+						Fatalf("doModuleInstallOrNothing(): 1 Error while creating symlink " + workDir + " pointing to " + absolutePath + " Error: " + err.Error())
 					}
 					//} else {
 					//Debugf("need to fetch Forge module " + moduleName + " in latest, because version " + fr.versionNumber + " will not be fetched already")
@@ -57,7 +78,6 @@ func doModuleInstallOrNothing(fm ForgeModule) {
 			}
 		} else {
 			if fm.cacheTTL > 0 {
-				lastCheckedFile := workDir + "-last-checked"
 				//Debugf("checking for " + lastCheckedFile)
 				if fileInfo, err := os.Lstat(lastCheckedFile); err == nil {
 					//Debugf("found " + lastCheckedFile + " with mTime " + fileInfo.ModTime().String())
@@ -69,20 +89,8 @@ func doModuleInstallOrNothing(fm ForgeModule) {
 						latestForgeModules.m[moduleName] = me.version
 						latestForgeModules.Unlock()
 
-						// check content of lastCheckedFile (which should be the Forge API response body) if the module is deprecated
-
-						if fi, err := os.Stat(lastCheckedFile); err == nil {
-							if fi.Size() < 1 {
-								Debugf("found empty file " + lastCheckedFile)
-								fr = queryForgeAPI(fm)
-							} else {
-								json, err := ioutil.ReadFile(lastCheckedFile)
-								if err != nil {
-									Fatalf("doModuleInstallOrNothing(): Error while reading Forge API result from file " + lastCheckedFile + err.Error())
-								}
-								_ = parseForgeAPIResult(string(json), fm)
-								return
-							}
+						if checkDeprecation(fm, lastCheckedFile) {
+							return
 						}
 
 					}
@@ -99,7 +107,7 @@ func doModuleInstallOrNothing(fm ForgeModule) {
 
 	} else if moduleVersion == "present" {
 		// ensure that a latest version this module exists
-		latestDir := config.ForgeCacheDir + moduleName + "-latest"
+		latestDir := filepath.Join(config.ForgeCacheDir, moduleName+"-latest")
 		if !isDir(latestDir) {
 			if _, ok := uniqueForgeModules[moduleName+"-latest"]; ok {
 				Debugf("we got " + fm.author + "-" + fm.name + "-" + fm.version + ", but no " + latestDir + " to use, but -latest is already being fetched.")
@@ -113,8 +121,12 @@ func doModuleInstallOrNothing(fm ForgeModule) {
 		Debugf("Nothing to do for module " + fm.author + "-" + fm.name + "-" + fm.version + ", because " + latestDir + " exists")
 	} else {
 		if !isDir(workDir) {
+			_ = queryForgeAPI(fm)
 			fr.needToGet = true
 		} else {
+			if !checkDeprecation(fm, lastCheckedFile) {
+				_ = queryForgeAPI(fm)
+			}
 			Debugf("Using cache for " + moduleName + " in version " + moduleVersion + " because " + workDir + " exists")
 			return
 		}
@@ -129,14 +141,14 @@ func doModuleInstallOrNothing(fm ForgeModule) {
 		} else {
 			// os.Readlink error is okay
 			versionDir, _ := os.Readlink(workDir)
-			if versionDir == config.ForgeCacheDir+moduleName+"-"+fr.versionNumber {
+			if versionDir == filepath.Join(config.ForgeCacheDir, moduleName+"-"+fr.versionNumber) {
 				Debugf("No reason to re-symlink again")
 			} else {
 				if isDir(workDir) {
 					Debugf("Trying to remove symlink: " + workDir)
 					_ = os.Remove(workDir)
 				}
-				versionDir = config.ForgeCacheDir + moduleName + "-" + fr.versionNumber
+				versionDir = filepath.Join(config.ForgeCacheDir, moduleName+"-"+fr.versionNumber)
 				absolutePath, err := filepath.Abs(versionDir)
 				if err != nil {
 					Fatalf("doModuleInstallOrNothing(): Error while resolving absolute file path for " + versionDir + " Error: " + err.Error())
@@ -153,11 +165,11 @@ func doModuleInstallOrNothing(fm ForgeModule) {
 }
 
 func queryForgeAPI(fm ForgeModule) ForgeResult {
-	baseURL := config.Forge.Baseurl
+	baseURL := config.ForgeBaseURL
 	if len(fm.baseURL) > 0 {
 		baseURL = fm.baseURL
 	}
-	url := baseURL + "/v3/modules/" + fm.author + "-" + fm.name
+	url := baseURL + "/v3/modules/" + fm.author + "-" + fm.name + "?exclude_fields=changelog+readme+license+releases"
 	req, err := http.NewRequest("GET", url, nil)
 	if err != nil {
 		Fatalf("queryForgeAPI(): Error creating GET request for Puppetlabs forge API" + err.Error())
@@ -198,7 +210,7 @@ func queryForgeAPI(fm ForgeModule) ForgeResult {
 		json := string(body)
 		fr := parseForgeAPIResult(json, fm)
 
-		lastCheckedFile := config.ForgeCacheDir + fm.author + "-" + fm.name + "-latest-last-checked"
+		lastCheckedFile := filepath.Join(config.ForgeCacheDir, fm.author+"-"+fm.name+"-latest-last-checked")
 		Debugf("writing last-checked file " + lastCheckedFile)
 		f, _ := os.Create(lastCheckedFile)
 		defer f.Close()
@@ -264,12 +276,15 @@ func parseForgeAPIResult(json string, fm ForgeModule) ForgeResult {
 
 // getMetadataForgeModule queries the configured Puppet Forge and return
 func getMetadataForgeModule(fm ForgeModule) ForgeModule {
-	baseURL := config.Forge.Baseurl
+	baseURL := config.ForgeBaseURL
 	if len(fm.baseURL) > 0 {
 		baseURL = fm.baseURL
 	}
 	url := baseURL + "/v3/releases/" + fm.author + "-" + fm.name + "-" + fm.version
 	req, err := http.NewRequest("GET", url, nil)
+	if err != nil {
+		Fatalf("getMetadataForgeModule(): Error while creating GET http request with url " + url + " Error: " + err.Error())
+	}
 	req.Header.Set("User-Agent", "https://github.com/xorpaul/g10k/")
 	req.Header.Set("Connection", "keep-alive")
 	proxyURL, err := http.ProxyFromEnvironment(req)
@@ -321,7 +336,7 @@ func extractForgeModule(wgForgeModule *sync.WaitGroup, file *io.PipeReader, file
 	before := time.Now()
 	fileReader, err := pgzip.NewReader(file)
 
-	unTar(fileReader, config.ForgeCacheDir+"/")
+	unTar(fileReader, config.ForgeCacheDir)
 
 	if err != nil {
 		Fatalf(funcName + "(): pgzip reader error for module " + fileName + " error:" + err.Error())
@@ -329,98 +344,12 @@ func extractForgeModule(wgForgeModule *sync.WaitGroup, file *io.PipeReader, file
 	defer fileReader.Close()
 
 	duration := time.Since(before).Seconds()
-	Verbosef("Extracting " + config.ForgeCacheDir + fileName + " took " + strconv.FormatFloat(duration, 'f', 5, 64) + "s")
+	Verbosef("Extracting " + filepath.Join(config.ForgeCacheDir, fileName) + " took " + strconv.FormatFloat(duration, 'f', 5, 64) + "s")
 	mutex.Lock()
 	ioForgeTime += duration
 	mutex.Unlock()
 }
 
-func unTar(r io.Reader, targetBaseDir string) {
-	funcName := funcName()
-	tarBallReader := tar.NewReader(r)
-	for {
-		header, err := tarBallReader.Next()
-		if err != nil {
-			if err == io.EOF {
-				break
-			}
-			Fatalf(funcName + "(): error while tar reader.Next() for io.Reader with targetBaseDir " + targetBaseDir + err.Error())
-		}
-
-		// get the individual filename and extract to the current directory
-		filename := header.Name
-		targetFilename := targetBaseDir + filename
-
-		switch header.Typeflag {
-		case tar.TypeDir:
-			// handle directory
-			//fmt.Println("Creating directory :", filename)
-			//err = os.MkdirAll(targetFilename, os.FileMode(header.Mode)) // or use 0755 if you prefer
-			err = os.MkdirAll(targetFilename, os.FileMode(0755)) // or use 0755 if you prefer
-
-			if err != nil {
-				Fatalf(funcName + "(): error while MkdirAll() " + filename + err.Error())
-			}
-
-			err = os.Chtimes(targetFilename, header.AccessTime, header.ModTime)
-
-			if err != nil {
-				Fatalf(funcName + "(): error while Chtimes() " + filename + err.Error())
-			}
-
-		case tar.TypeReg:
-			// handle normal file
-			//fmt.Println("Untarring :", filename)
-			writer, err := os.Create(targetFilename)
-
-			if err != nil {
-				Fatalf(funcName + "(): error while Create() " + filename + err.Error())
-			}
-
-			io.Copy(writer, tarBallReader)
-			err = os.Chmod(targetFilename, os.FileMode(header.Mode))
-
-			if err != nil {
-				Fatalf(funcName + "(): error while Chmod() " + filename + err.Error())
-			}
-
-			err = os.Chtimes(targetFilename, header.AccessTime, header.ModTime)
-			if err != nil {
-				Fatalf(funcName + "(): error while Chtimes() " + filename + err.Error())
-			}
-
-			writer.Close()
-
-		case tar.TypeSymlink:
-			if os.Symlink(header.Linkname, targetFilename) != nil {
-				Fatalf(funcName + "(): error while creating symlink " + targetFilename + " pointing to " + header.Linkname + " Error: " + err.Error())
-			}
-
-		case tar.TypeLink:
-			if os.Link(header.Linkname, targetFilename) != nil {
-				Fatalf(funcName + "(): error while creating hardlink " + targetFilename + " pointing to " + header.Linkname + " Error: " + err.Error())
-			}
-
-		// Skip pax_global_header with the commit ID this archive was created from
-		case tar.TypeXGlobalHeader:
-			continue
-
-		default:
-			Fatalf(funcName + "(): Unable to untar type: " + string(header.Typeflag) + " in file " + filename)
-		}
-	}
-	// tarball produced by git archive has trailing nulls in the stream which are not
-	// read by the module, when removed this can cause the git archive to hang trying
-	// to output the nulls into a full pipe buffer, avoid this by discarding the rest
-	// until the stream ends.
-	buf := make([]byte, 4096)
-	nread, err := r.Read(buf)
-	for nread > 0 && err == nil {
-		Debugf(fmt.Sprintf("Discarded %d bytes of trailing data from tar", nread))
-		nread, err = r.Read(buf)
-	}
-}
-
 func downloadForgeModule(name string, version string, fm ForgeModule, retryCount int) {
 	funcName := funcName()
 	var wgForgeModule sync.WaitGroup
@@ -428,16 +357,19 @@ func downloadForgeModule(name string, version string, fm ForgeModule, retryCount
 	extractR, extractW := io.Pipe()
 	saveFileR, saveFileW := io.Pipe()
 
-	//url := "https://forgeapi.puppetlabs.com/v3/files/puppetlabs-apt-2.1.1.tar.gz"
+	//url := "https://forgeapi.puppet.com/v3/files/puppetlabs-apt-2.1.1.tar.gz"
 	fileName := name + "-" + version + ".tar.gz"
 
-	if !isDir(config.ForgeCacheDir + name + "-" + version) {
-		baseURL := config.Forge.Baseurl
+	if !isDir(filepath.Join(config.ForgeCacheDir, name+"-"+version)) {
+		baseURL := config.ForgeBaseURL
 		if len(fm.baseURL) > 0 {
 			baseURL = fm.baseURL
 		}
 		url := baseURL + "/v3/files/" + fileName
 		req, err := http.NewRequest("GET", url, nil)
+		if err != nil {
+			Fatalf("getMetadataForgeModule(): Error while creating GET http request with url " + url + " Error: " + err.Error())
+		}
 		req.Header.Set("User-Agent", "https://github.com/xorpaul/g10k/")
 		req.Header.Set("Connection", "close")
 		proxyURL, err := http.ProxyFromEnvironment(req)
@@ -462,14 +394,15 @@ func downloadForgeModule(name string, version string, fm ForgeModule, retryCount
 			wgForgeModule.Add(1)
 			go func() {
 				defer wgForgeModule.Done()
-				Debugf(funcName + "(): Trying to create " + config.ForgeCacheDir + fileName)
-				out, err := os.Create(config.ForgeCacheDir + fileName)
+				targetFileName := filepath.Join(config.ForgeCacheDir, fileName)
+				Debugf(funcName + "(): Trying to create " + targetFileName)
+				out, err := os.Create(targetFileName)
 				if err != nil {
-					Fatalf(funcName + "(): Error while creating file for Forge module " + config.ForgeCacheDir + fileName + " Error: " + err.Error())
+					Fatalf(funcName + "(): Error while creating file for Forge module " + targetFileName + " Error: " + err.Error())
 				}
 				defer out.Close()
 				io.Copy(out, saveFileR)
-				Debugf(funcName + "(): Finished creating " + config.ForgeCacheDir + fileName)
+				Debugf(funcName + "(): Finished creating " + targetFileName)
 			}()
 			wgForgeModule.Add(1)
 			go extractForgeModule(&wgForgeModule, extractR, fileName)
@@ -483,17 +416,20 @@ func downloadForgeModule(name string, version string, fm ForgeModule, retryCount
 				defer extractW.Close()
 				defer saveFileW.Close()
 
-				var mw io.Writer
 				// build the multiwriter for all the pipes
-				mw = io.MultiWriter(extractW, saveFileW)
+				mw := io.MultiWriter(extractW, saveFileW)
 
 				// copy the data into the multiwriter
 				if _, err := io.Copy(mw, resp.Body); err != nil {
 					Fatalf("Error while writing to MultiWriter " + err.Error())
 				}
 			}()
+		} else if strings.TrimSpace(resp.Status) == "404 Not Found" {
+			Fatalf("Received 404 from Forge using URL " + url +
+				"\nCheck if the module name '" + fm.author + "-" + fm.name + "' and version '" + version + "' really exist" +
+				"\nUsed in Puppet environment '" + fm.sourceBranch + "'")
 		} else {
-			Fatalf(funcName + "(): Unexpected response code while GETing " + url + " " + resp.Status)
+			Fatalf("Unexpected response code while GETing " + url + " " + resp.Status)
 		}
 	} else {
 		Debugf("Using cache for Forge module " + name + " version: " + version)
@@ -507,8 +443,8 @@ func downloadForgeModule(name string, version string, fm ForgeModule, retryCount
 				Fatalf("downloadForgeModule(): giving up for Puppet module " + name + " version: " + version)
 			}
 			Warnf("Retrying...")
-			purgeDir(config.ForgeCacheDir+fileName, "downloadForgeModule()")
-			purgeDir(strings.Replace(config.ForgeCacheDir+fileName, ".tar.gz", "/", -1), "downloadForgeModule()")
+			purgeDir(filepath.Join(config.ForgeCacheDir, fileName), "downloadForgeModule()")
+			purgeDir(strings.Replace(filepath.Join(config.ForgeCacheDir, fileName), ".tar.gz", "/", -1), "downloadForgeModule()")
 			// retry if hash sum mismatch found
 			downloadForgeModule(name, version, fm, retryCount-1)
 		}
@@ -632,7 +568,7 @@ func doForgeModuleIntegrityCheck(m ForgeModule) bool {
 	md5R, md5W := io.Pipe()
 	sha256R, sha256W := io.Pipe()
 	var calculatedArchiveSize int64
-	fileName := config.ForgeCacheDir + m.author + "-" + m.name + "-" + m.version + ".tar.gz"
+	fileName := filepath.Join(config.ForgeCacheDir, m.author+"-"+m.name+"-"+m.version+".tar.gz")
 
 	// md5 sum
 	wgCheckSum.Add(1)
@@ -738,12 +674,13 @@ func syncForgeToModuleDir(name string, m ForgeModule, moduleDir string, correspo
 	mutex.Unlock()
 	moduleName := m.author + "-" + m.name
 	//Debugf("m.name " + m.name + " m.version " + m.version + " moduleName " + moduleName)
-	targetDir := normalizeDir(moduleDir + m.name)
+	targetDir := filepath.Join(moduleDir, m.name)
+	metadataFile := filepath.Join(targetDir, "metadata.json")
 	if m.version == "present" {
-		if fileExists(targetDir + "metadata.json") {
-			Debugf("Nothing to do, found existing Forge module: " + targetDir + "metadata.json")
+		if fileExists(metadataFile) {
+			Debugf("Nothing to do, found existing Forge module: " + targetDir)
 			if check4update {
-				me := readModuleMetadata(targetDir + "metadata.json")
+				me := readModuleMetadata(metadataFile)
 				latestForgeModules.RLock()
 				check4ForgeUpdate(m.name, me.version, latestForgeModules.m[moduleName])
 				latestForgeModules.RUnlock()
@@ -755,8 +692,8 @@ func syncForgeToModuleDir(name string, m ForgeModule, moduleDir string, correspo
 
 	}
 	if isDir(targetDir) {
-		if fileExists(targetDir + "metadata.json") {
-			me := readModuleMetadata(targetDir + "metadata.json")
+		if fileExists(metadataFile) {
+			me := readModuleMetadata(metadataFile)
 			if m.version == "latest" {
 				//fmt.Println(latestForgeModules)
 				//fmt.Println("checking latestForgeModules for key", moduleName)
@@ -783,18 +720,22 @@ func syncForgeToModuleDir(name string, m ForgeModule, moduleDir string, correspo
 			createOrPurgeDir(targetDir, " targetDir for module "+m.name+" with missing metadata.json")
 		}
 	}
-	workDir := normalizeDir(config.ForgeCacheDir + moduleName + "-" + m.version)
-	if !isDir(workDir) {
+	workDir := normalizeDir(filepath.Join(config.ForgeCacheDir, moduleName+"-"+m.version))
+	resolvedWorkDir, err := filepath.EvalSymlinks(workDir)
+	if err != nil {
+		Fatalf(funcName + "(): Failed to resolve possivle symlink " + workDir + " Error: " + err.Error())
+	}
+	if !isDir(resolvedWorkDir) {
 		if config.UseCacheFallback {
-			Warnf("Failed to use " + workDir + " Trying to use latest cached version of module " + moduleName)
-			workDir = getLatestCachedModule(m)
+			Warnf("Failed to use " + resolvedWorkDir + " Trying to use latest cached version of module " + moduleName)
+			resolvedWorkDir = getLatestCachedModule(m)
 		} else {
-			Fatalf(funcName + "(): Forge module not found in dir: " + workDir)
+			Fatalf(funcName + "(): Forge module not found in dir: " + resolvedWorkDir)
 		}
 	}
 
-	if !isDir(workDir) {
-		Fatalf(funcName + "(): Forge module not found in dir: " + workDir)
+	if !isDir(resolvedWorkDir) {
+		Fatalf(funcName + "(): Forge module not found in dir: " + resolvedWorkDir)
 	}
 
 	Infof("Need to sync " + targetDir)
@@ -808,16 +749,16 @@ func syncForgeToModuleDir(name string, m ForgeModule, moduleDir string, correspo
 		} else {
 			Fatalf(funcName + "(): Error while os.Stat file " + targetDir)
 		}
-		if fileInfo, err := os.Stat(workDir); err == nil {
+		if fileInfo, err := os.Stat(resolvedWorkDir); err == nil {
 			if fileInfo.Sys() != nil {
 				workDirDevice = uint64(fileInfo.Sys().(*syscall.Stat_t).Dev)
 			}
 		} else {
-			Fatalf(funcName + "(): Error while os.Stat file " + workDir)
+			Fatalf(funcName + "(): Error while os.Stat file " + resolvedWorkDir)
 		}
 
-		if targetDirDevice != workDirDevice {
-			Fatalf("Error: Can't hardlink Forge module files over different devices. Please consider changing the cachdir setting. ForgeCachedir: " + config.ForgeCacheDir + " target dir: " + targetDir)
+		if targetDirDevice != workDirDevice && !usemove {
+			Fatalf("Error: Can't hardlink Forge module files over different devices. Please consider changing the cachedir setting. ForgeCachedir: " + config.ForgeCacheDir + " target dir: " + targetDir)
 		}
 
 		mutex.Lock()
@@ -827,22 +768,34 @@ func syncForgeToModuleDir(name string, m ForgeModule, moduleDir string, correspo
 		}
 		needSyncForgeCount++
 		mutex.Unlock()
-		hardlink := func(path string, info os.FileInfo, err error) error {
-			if filepath.Base(path) != filepath.Base(workDir) { // skip the root dir
-				target, err := filepath.Rel(workDir, path)
-				if err != nil {
-					Fatalf(funcName + "(): Can't make " + path + " relative to " + workDir + " Error: " + err.Error())
-				}
+		destination := func(path string, info os.FileInfo, err error) error {
+			if err != nil {
+				Fatalf(funcName + "(): Error while calling generic func() Error " + err.Error())
+			}
+			target, err := filepath.Rel(resolvedWorkDir, path)
+			if err != nil {
+				Fatalf(funcName + "(): Can't make " + path + " relative to " + resolvedWorkDir + " Error: " + err.Error())
+			}
 
-				if info.IsDir() {
-					//Debugf(funcName + "() Trying to mkdir " + targetDir + target)
-					err = os.Mkdir(targetDir+"/"+target, os.FileMode(0755))
+			if info.IsDir() {
+				if target != "." { // skip the root dir
+					//Debugf(funcName + "() Trying to mkdir " + filepath.Join(targetDir, target))
+					err = os.Mkdir(filepath.Join(targetDir, target), os.FileMode(0755))
 					if err != nil {
 						Fatalf(funcName + "(): error while Mkdir() " + targetDir + "/" + target + " Error: " + err.Error())
 					}
+				}
+			} else {
+				if usemove {
+					//Debugf(funcName + "() Trying to helper.moveFile " + path + " to " + targetDir + target)
+					// deleteSourceFileToggle is set to false as we delete the source file later in the main() anyway after the sync completes
+					err = moveFile(path, filepath.Join(targetDir, target), false)
+					if err != nil {
+						Fatalf(funcName + "(): Failed to helper.moveFile " + path + " to " + targetDir + "/" + target + " Error: " + err.Error())
+					}
 				} else {
-					//Debugf(funcName + "() Trying to hardlink " + path + " to " + targetDir + target)
-					err = os.Link(path, targetDir+"/"+target)
+					//Debugf(funcName + "() Trying to hardlink " + path + " to " + filepath.Join(targetDir, target))
+					err = os.Link(path, filepath.Join(targetDir, target))
 					if err != nil {
 						Fatalf(funcName + "(): Failed to hardlink " + path + " to " + targetDir + "/" + target + " Error: " + err.Error())
 					}
@@ -850,10 +803,11 @@ func syncForgeToModuleDir(name string, m ForgeModule, moduleDir string, correspo
 			}
 			return nil
 		}
+
 		c := make(chan error)
-		Debugf(funcName + "() filepath.Walk'ing directory " + workDir)
+		Debugf(funcName + "() filepath.Walk'ing directory " + resolvedWorkDir)
 		before := time.Now()
-		go func() { c <- filepath.Walk(workDir, hardlink) }()
+		go func() { c <- filepath.Walk(resolvedWorkDir, destination) }()
 		<-c // Walk done
 		duration := time.Since(before).Seconds()
 		mutex.Lock()
@@ -867,10 +821,10 @@ func syncForgeToModuleDir(name string, m ForgeModule, moduleDir string, correspo
 func getLatestCachedModule(m ForgeModule) string {
 	latest := "//"
 	version := "latest"
-	latestDir := config.ForgeCacheDir + m.author + "-" + m.name + "-latest"
+	latestDir := filepath.Join(config.ForgeCacheDir, m.author+"-"+m.name+"-latest")
 	if !isDir(latestDir) {
 
-		globPath := config.ForgeCacheDir + m.author + "-" + m.name + "-*"
+		globPath := filepath.Join(config.ForgeCacheDir, m.author+"-"+m.name+"-*")
 		Debugf("Glob'ing with path " + globPath)
 		matches, err := filepath.Glob(globPath)
 		if len(matches) == 0 {
@@ -885,7 +839,7 @@ func getLatestCachedModule(m ForgeModule) string {
 			if isDir(m) {
 				Debugf("Comparing " + latest + " < " + m)
 				if latest < m {
-					Debugf("Setting " + latest + " to " + m)
+					Debugf("Setting latest to " + m)
 					latest = m
 				}
 			}
@@ -907,7 +861,7 @@ func getLatestCachedModule(m ForgeModule) string {
 		}
 
 		version = strings.Split(versionDir, m.author+"-"+m.name+"-")[1]
-		latest = config.ForgeCacheDir + m.author + "-" + m.name + "-" + version
+		latest = filepath.Join(config.ForgeCacheDir, m.author+"-"+m.name+"-"+version)
 	}
 
 	if latest == "//" {
